#include <gmsh.h>
#include <iostream>
#include <sstream>
#include <map>
#include <Eigen/Dense>
#include <omp.h>

// Each element has all its information gathered as in this structure.
struct elementStruct{
    std::size_t tag;        // tag of the element.
    double bcValue [2];     // values of [phi, phi_n].
    size_t nodeTags [2];    // tags of the [first, second] nodes of the element.
    size_t midNodeTag;      // tag of the mid element node (or 0 if 2-node element).
    double length;          // length of the element.
    double x1, y1, x2, y2;  // coordinates of the first/second nodes of the element.
    double midX, midY;      // mid point coordinates.
    double deltaX, deltaY;  // x and y variations (sign matters).
};

/*
 *  Main
 */

// Fills the "electrostaticPressure" map that gives the electrostatic pressure associated with each (boundary) element tag.
// "nbViews" is the counter of views (useful if other views have to be generated than with this solver). "nodalDisplacements" is
// a map associating the (x,y) displacement to a serie of node tags (if the position of the nodes are not striclty the one
// extracted from the .geo file). "postProcessing" indicates whether the post-processing should be done ('true') or not
// ('false'). "iteration" represents the number of times the solver has been called (some information is printed on the first
// iteration). "untangleMesh" unstangles every BEM domain if set to 'true' (in case the "nodalDisplacements" map contains
// non-zero displacements).
void solverBEM(std::map<int, double> &electrostaticPressure,
               int & nbViews,
               std::map<int,std::pair<double,double>> & nodalDisplacements,
               const bool postProcessing,
               const int iteration,
               const bool untangleMesh);

// Fills the "electrostaticPressure" map. The "nodalDisplacements", "postProcessing" and "iteration" arguments are described in
// the description of the function "solverBEM". "domainName" is used for extracting everything in the appropriate independant BEM
// domain, "phiViewTag" and "electricFieldViewTag" are the view tag of the potential and of the electric field respectively.
// "computePhi" and "computeElectricField" indicates whether the potential and/or the electric field should be computed ('true')
// or not ('false').
void singleDomainSolverBEM(std::map<int, double> &electrostaticPressure,
                           const int nbViews,
                           std::map<int,std::pair<double,double>> & nodalDisplacements,
                           const bool postProcessing,
                           const int iteration,
                           const std::string domainName,
                           const int phiViewTag,
                           const int electricFieldViewTag,
                           const bool computePhi,
                           const bool computeElectricField);

/*
 *  Boundary element
 */

// Fills the "boundaryConditions" map associating to each physical group (a string) the type of boundary condition ('true':
// dirichlet, 'false': neumann) and its value (if any). "domainName" is the name of the current BEM domain.
void getBC(std::map<std::string, std::pair<bool, double>> &boundaryConditions,
           const std::string domainName);

// Fills the "elementVector" vector storing the boundary elements (see "elementStruct" in the 'functions.hpp' file) of the
// current BEM domain, with the appropriate information for each element (the bcValue will be computed later). Also returns
// "domainPhysicalGroupDim" and "domainPhysicalGroupTag", the dimension and the tag (respectivelly) of the BEM domain of
// interest. "domainName" is the name of the current BEM domain. "nodalDisplacements" is a map associating the (x,y) displacement
// to a serie of node tags. Returns the number of 'dirichlet elements' (useful for the construction/resolution of the system).
std::size_t fillElementVector(std::vector<elementStruct> &elementVector,
                              int &domainPhysicalGroupDim,
                              int &domainPhysicalGroupTag,
                              std::map<int, std::pair<double, double>> &nodalDisplacements,
                              const std::string domainName);

// Computes the useful data of a given element and stores it in "element". "nodeIndinces" is a map associating the index of any
// node tag for accessing its coordinates in the "nodeCoords" vector. "nodalDisp1" and "nodalDisp2" are pairs representing the
// potential displacement of the first and second node (respectivelly) of the element. If an element has 3 nodes, the first and
// second nodes are considered to be the extremities of that element.
void computeGeometricParam(elementStruct &element,
                           std::map<std::size_t, std::size_t> &nodeIndices,
                           const std::vector<double> &nodeCoords,
                           const std::pair<double, double> &nodalDisp1,
                           const std::pair<double, double> &nodalDisp2);

// Fills the "A", "B", "c" matrices (/vector) that represent the linear system. "elementVector" is a vector storing the boundary
// elements (see "elementStruct" in the 'functions.hpp' file) of the current BEM domain. "nbDirichletElements" is the number of
// 'dirichlet elements'. "solutionType" indicates whether the elements of the matrices are computed analytically ('true') or
// numerically ('false'). "integrationType" is the type of numerical integration used (in case of numerical computation).
void fillSystem(Eigen::MatrixXd &A,
                Eigen::MatrixXd &B,
                Eigen::MatrixXd &c,
                const std::vector<elementStruct> &elementVector,
                const std::size_t nbDirichletElements,
                const std::string integrationType,
                const bool solutionType);

// Fills the "electrostaticPressure" map associating each (boundary) element tag with its electrostatic pressure. Also sets the
// unknown of each element (either the potential of the electric field) of the "elementVector" vector storing the boundary
// elements (see "elementStruct" in the 'functions.hpp' file) of the current BEM domain. "x" is the solution of the linear
// system. "nbDirichletElements" is the number of 'dirichlet elements', "domainName" is the name of the considered BEM domain.
void computeElectrostaticPressure(std::map<int, double> &electrostaticPressure,
                                  std::vector<elementStruct> &elementVector,
                                  const Eigen::MatrixXd &x,
                                  const std::size_t nbDirichletElements,
                                  const std::string domainName);

/*
 * Post processing
 */

// Retrieves every usefull information from the inside domain for the post-processing. "domainNodeTags" and "domainNodeCoords"
// are vectors containing all the node tags and coordinates (respectivelly) of the domain. "domainNodeIndices" is a map
// associating the index of each node tag for accessing the "domainNodeCoords" vector. "domainEntityTags" is a vector containing
// the tag of each entity of the domain. "domainElementTypes" stores the types of the elements of the the domain. The first
// dimension of this vector represents the different entities and the second, the different element types. "domainElementTags"
// stores the tags of the elements of the domain. The first dimension of this vector represents the different entities, the
// second, the different element types and the third, the different element tags. The "domainElementNodeTags" map associates to
// each element tag a vector of node tags representing the nodes of that element. "domainPhysicalGroupTag" and
// "domainPhysicalGroupDim" are the dimension and tag (respectivelly) of the BEM domain of interrest.
void getModel(std::vector<std::size_t> &domainNodeTags,
              std::map<std::size_t, std::size_t> &domainNodeIndices,
              std::vector<double> &domainNodeCoords,
              std::vector<int> &domainEntityTags,
              std::vector<std::vector<int>> &domainElementTypes,
              std::vector<std::vector<std::vector<std::size_t>>> &domainElementTags,
              std::map<std::size_t, std::vector<std::size_t>> &domainElementNodeTags,
              const int domainPhysicalGroupDim,
              const int domainPhysicalGroupTag);

// From the "elementVector" vector containing the information of all elements, constructs the "boundaryNodeToElements" map
// associating to each boundary node tag, a vector of element tags of the contiguous elements. The vector of element tags
// contains 2 element excepts in case of a mid element node (for higher order meshes), where only one element tag is stored.
void computeNodeToElements(std::map<std::size_t, std::vector<std::size_t>> &boundaryNodeToElements,
                           const std::vector<elementStruct> &elementVector);

// Computes and stores the potential of each interior node in the "internalNodalPhi" (it can be accessed via the tag of the
// node). "domainNodeTags" and "domainNodeCoords" are vectors containing all the node tags and coordinates (respectivelly) of the
// domain. "domainNodeIndices" is a map associating the index of each node tag for accessing the "domainNodeCoords" vector.
// "elementVector" is a vector storing the boundary elements (see "elementStruct" in the 'functions.hpp' file) of the current BEM
// domain. "solutionType" indicates whether the elements of the matrices are computed analytically ('true') or numerically
// ('false'). "integrationType" is the type of numerical integration used (in case of numerical computation).
// "boundaryNodeToElements" is a map associating to each boundary node tag, a vector of element tags of the contiguous elements.
void computeInternalPhi(std::map<int, double> &internalNodalPhi,
                        const std::vector<double> &domainNodeCoords,
                        const std::vector<std::size_t> &domainNodeTags,
                        std::map<std::size_t, std::size_t> &domainNodeIndices,
                        const std::vector<elementStruct> &elementVector,
                        const std::string integrationType,
                        std::map<std::size_t, std::vector<std::size_t>> &boundaryNodeToElements,
                        const bool &solutionType);

// Binds the potential at each node to the right elements. "phi" is a data structure storing the potiential of each node (fourth
// dimension) of each element tag (third dimension) of each element type (second dimension) of each entity tag (first dimension).
// "domainElementTags" stores the tag of each element (third dimension) of each element type (second dimension) of each entity
// tag (first dimension). "domainElementTypes" stores the type of each element (second dimension) of each entity (first
// dimension). The "domainElementNodeTags" map associates to each element tag a vector of node tags representing the nodes of
// that element. "elementVector" is a vector storing the boundary elements (see "elementStruct" in the 'functions.hpp' file) of
// the current BEM domain. "internalNodalPhi" is a map associating the value of the potential to each node tag of the domain
// (boundaries excuded). "domainPhysicalGroupTag" is the tag of the physical domain of interrest. "boundaryNodeToElements" is a
// map associating to each boundary node tag, a vector of element tags of the contiguous elements.
void elementNodeData(std::vector<std::vector<std::vector<std::vector<double>>>> &phi,
                     const std::vector<std::vector<std::vector<size_t>>> &domainElementTags,
                     const std::vector<std::vector<int>> &domainElementTypes,
                     std::map<std::size_t, std::vector<std::size_t>> &domainElementNodeTags,
                     const std::vector<elementStruct> &elementVector,
                     std::map<int, double> &internalNodalPhi,
                     const int domainPhysicalGroupTag,
                     std::map<std::size_t, std::vector<std::size_t>> &boundaryNodeToElements);

// Initializes the "phi" data structure in order for it to have the proper dimensions. It is a data structure storing the
// potiential of each node (fourth dimension) of each element tag (third dimension) of each element type (second dimension) of
// each entity tag (first dimension). The "position" map associates to each of the element tag of the domain the 3 indices for
// accessing it in the "domainElementTags" or "phi" data structure. "domainElementTags" stores the tag of each element (third
// dimension) of each element type (second dimension) of each entity tag (first dimension). "domainElementTypes" stores the type
// of each element (second dimension) of each entity (first dimension).
void phiInitialization(std::vector<std::vector<std::vector<std::vector<double>>>> &phi,
                       std::map<std::size_t, std::vector<std::size_t>> &position,
                       const std::vector<std::vector<int>> &domainElementTypes,
                       const std::vector<std::vector<std::vector<size_t>>> &domainElementTags);

// Fills the "domainBoundaryElementTags" vector containing the tag of every elements possessing at least one node in common with
// the boundary. "domainElementTags" stores the tag of each element (third dimension) of each element type (second dimension) of
// each entity tag (first dimension). "domainElementTypes" stores the type of each element (second dimension) of each entity
// (first dimension). "elementVector" is a vector storing the boundary elements (see "elementStruct" in the 'functions.hpp' file)
// of the current BEM domain. "domainPhysicalGroupTag" is the tag of the physical domain of interrest.
void boundaryElementTags(std::vector<std::size_t> &domainBoundaryElementTags,
                         const std::vector<std::vector<int>> &domainElementTypes,
                         const std::vector<std::vector<std::vector<std::size_t>>> &domainElementTags,
                         const std::vector<elementStruct> &elementVector,
                         const int domainPhysicalGroupTag);

// Associates the potential to the nodes of each elements (if not already done) in the "phi" data structure. It is a data
// structure storing the potiential of each node (fourth dimension) of each element tag (third dimension) of each element type
// (second dimension) of each entity tag (first dimension). "domainElementTags" stores the tag of each element (third dimension)
// of each element type (second dimension) of each entity tag (first dimension). The "domainElementNodeTags" map associates to
// each element tag a vector of node tags representing the nodes of that element. "internalNodalPhi" is a map associating the
// value of the potential to each node tag of the domain (boundaries excuded).
void elementNodalPhiAssociation(std::vector<std::vector<std::vector<std::vector<double>>>> &phi,
                                const std::vector<std::vector<std::vector<std::size_t>>> &domainElementTags,
                                std::map<std::size_t, std::vector<std::size_t>> &domainElementNodeTags,
                                std::map<int, double> &internalNodalPhi);


// Computes the potential (if "computePhi" is set to 'true') and the electric field (if "computeElectricField" is set to 'true')
// and stores them in the "phi" and "electricField" datastructures respectivelly. These assign the value of the potential /
// electric field to each element tag (third dimension) of each element type (second dimension) of each entity tag (first
// dimension) of the domain. The fourth dimension is either of size 1 (for the potential) or 3 (for the electric field).
// "domainEntityTags" is a vector containing the tag of each entity of the domain. "domainElementTypes" stores the types of the
// elements of the the domain. The first dimension of this vector represents the different entities and the second, the different
// element types. "domainElementTags" stores the tags of the elements of the domain. The first dimension of this vector
// represents the different entities, the second, the different element types and the third, the different element tags.
// "elementVector" is a vector storing the boundary elements (see "elementStruct" in the 'functions.hpp' file) of the current BEM
// domain. "solutionType" indicates whether the elements of the matrices are computed analytically ('true') or numerically
// ('false'). "integrationType" is the type of numerical integration used (in case of numerical computation).
void computeElementData(std::vector<std::vector<std::vector<std::vector<double>>>> &phi,
                        std::vector<std::vector<std::vector<std::vector<double>>>> &electricField,
                        const std::vector<std::vector<std::vector<std::size_t>>> &domainElementTags,
                        const std::vector<std::vector<int>> &domainElementTypes,
                        const std::vector<int> &domainEntityTags,
                        const std::vector<elementStruct> &elementVector,
                        const bool solutionType,
                        const std::string integrationType,
                        const bool computePhi,
                        const bool computeElectricField);

// Deals with the display in gmsh. "viewTag" is the view tag of the data. "modelName" is the name of the model. "fileName" is the
// name of the output file. "dataType" is either "ElementData" or "ElementNodeData" (which are the supported type of this
// application). "domainElementTags" stores the tags of the elements of the domain. The first dimension of this vector represents
// the different entities, the second, the different element types and the third, the different element tags. "data" is a data
// structure storing the data to be displayed. The first three dimensions follow the same structure as the one of the
// "elementNodeTags" and the fourth one is contains what needs to be displayed. "viewIndex" if the index of the view.
void displayData(const int viewTag,
                 const std::string modelName,
                 const std::string fileName,
                 const std::string dataType,
                 const std::vector<std::vector<std::vector<std::size_t>>> &domainElementTags,
                 const std::vector<std::vector<std::vector<std::vector<double>>>> &data,
                 const int viewIndex);

/*
 *  Verification
 */

// Displays the time taken between the values of "start" and "end" (in seconds) if "showTime" is set to 'true'. "functionName" is
// the name of the function that has been timed.
void displayTime(const double start,
                 const double end,
                 std::string functionName, const bool showTime);

// Displays the different errors induced by the boundary element method compared with a simple analytical solution on a certain
// model geometry ("BEM_convergence.geo"). "domainEntityTags" is a vector containing the tag of each entity of the domain.
// "domainElementTypes" stores the types of the elements of the the domain. The first dimension of this vector represents the
// different entities and the second, the different element types. "domainElementTags" stores the tags of the elements of the
// domain. The first dimension of this vector represents the different entities, the second, the different element types and the
// third, the different element tags. "elementalPhi" and "elementalElectricField" are the used data for comparing numerical and
// analytical solutions. The convergence function can only be used with the "ElementData" data type.
void convergenceFun(const std::vector<int> &domainEntityTags,
                    const std::vector<std::vector<int>> &domainElementTypes,
                    const std::vector<std::vector<std::vector<size_t>>> &domainElementTags,
                    const std::vector<std::vector<std::vector<std::vector<double>>>> &elementalPhi,
                    const std::vector<std::vector<std::vector<std::vector<double>>>> &elementalElectricField);

/*
 *  Computations
 */

// Computes the 'h' element analytically. "x" and "y" are the coordinates at which 'h' should be computed. "element" is the
// element 'j' while will helps to compute 'h'.
double computeAnalyticalH(const double x,
                          const double y,
                          const elementStruct &element);

// Computes the 'g' element analytically. "x" and "y" are the coordinates at which 'g' should be computed. "element" is the
// element 'j' while will helps to compute 'g'.
double computeAnalyticalG(const double x,
                          const double y,
                          const elementStruct &element);

// Computes the 'hx' component of the gradient analytically. "x" and "y" are the coordinates at which 'gradHx' should be
// computed. "element" is the element 'j' while will helps to compute 'gradHx'.
double computeAnalyticalGradHx(const double x,
                               const double y,
                               const elementStruct &element);

// Computes the 'hy' component of the gradient analytically. "x" and "y" are the coordinates at which 'gradHy' should be
// computed. "element" is the element 'j' while will helps to compute 'gradHy'.
double computeAnalyticalGradHy(const double x,
                               const double y,
                               const elementStruct &element);

// Computes the 'gx' component of the gradient analytically. "x" and "y" are the coordinates at which 'gradGx' should be
// computed. "element" is the element 'j' while will helps to compute 'gradGx'.
double computeAnalyticalGradGx(const double x,
                               const double y,
                               const elementStruct &element);

// Computes the 'gy' component of the gradient analytically. "x" and "y" are the coordinates at which 'gradGy' should be
// computed. "element" is the element 'j' while will helps to compute 'gradGy'.
double computeAnalyticalGradGy(const double x,
                               const double y,
                               const elementStruct &element);

// Computes the 'g' element numerically. "x" and "y" are the coordinates at which 'g' should be computed. "element" is the
// element 'j' while will helps to compute 'g'. "integrationType" is the type of numerical integration used (in case of numerical
// computation).
double computeNumericalG(const double x,
                         const double y,
                         const elementStruct &element,
                         const std::string integrationType);

// Computes the 'hx' component of the gradient numerically. "x" and "y" are the coordinates at which 'gradHx' should be computed.
// "element" is the element 'j' while will helps to compute 'gradHx'. "integrationType" is the type of numerical integration used
// (in case of numerical computation).
double computeNumericalGradHx(const double x,
                              const double y,
                              const elementStruct &element,
                              const std::string integrationType);

// Computes the 'hy' component of the gradient numerically. "x" and "y" are the coordinates at which 'gradHy' should be computed.
// "element" is the element 'j' while will helps to compute 'gradHy'. "integrationType" is the type of numerical integration used
// (in case of numerical computation).
double computeNumericalGradHy(const double x,
                              const double y,
                              const elementStruct &element,
                              const std::string integrationType);

// Computes the 'gx' component of the gradient numerically. "x" and "y" are the coordinates at which 'gradGx' should be computed.
// "element" is the element 'j' while will helps to compute 'gradGx'. "integrationType" is the type of numerical integration used
// (in case of numerical computation).
double computeNumericalGradGx(const double x,
                              const double y,
                              const elementStruct &element,
                              const std::string integrationType);

// Computes the 'gy' component of the gradient numerically. "x" and "y" are the coordinates at which 'gradGy' should be computed.
// "element" is the element 'j' while will helps to compute 'gradGy'. "integrationType" is the type of numerical integration used
// (in case of numerical computation).
double computeNumericalGradGy(const double x,
                              const double y,
                              const elementStruct &element,
                              const std::string integrationType);